/* Do not modify this file. Changes will be overwritten.                      */
/* Generated automatically by the ASN.1 to Wireshark dissector compiler       */
/* packet-tetra.h                                                             */
/* asn2wrs.py -u -L -p tetra -c ./tetra.cnf -s ./packet-tetra-template -D . -O ../.. tetra.asn */

/* packet-tetra.h
 * Routines for TETRA packet dissection
 *
 * Copyright (c) 2007 - 2011 Professional Mobile Communication Research Group,
 *    Beijing Institute of Technology, China
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * REF: ETSI EN 300 392-2 V3.2.1
 */

#ifndef PACKET_TETRA_H
#define PACKET_TETRA_H

enum {
	TETRA_CHAN_AACH		= 1,
	TETRA_CHAN_SCH_F	= 2,
	TETRA_CHAN_SCH_D	= 3,
	TETRA_CHAN_BSCH		= 5,
	TETRA_CHAN_BNCH		= 6,
	TETRA_CHAN_TCH_F	= 7,
	TETRA_CHAN_TCH_H	= 8,
	TETRA_CHAN_TCH_2_4	= 9,
	TETRA_CHAN_TCH_4_8	= 10,
	TETRA_CHAN_STCH		= 11,
	TETRA_CHAN_SCH_HU	= 15
};

enum {
	TETRA_UPLINK,
	TETRA_DOWNLINK
};

void tetra_dissect_pdu(int channel_type, int dir, tvbuff_t *pdu, proto_tree *head, packet_info *pinfo);

/*#include "packet-tetra-exp.h"*/
#endif  /* PACKET_TETRA_H */
