/* Do not modify this file. */
/* It was automatically generated by ../../tools/process-x11-xcb.pl
   using xcbproto version xcb-proto-1.14-3-g7d58eed */
/*
 * Copyright 2008, 2009, 2013, 2014 Open Text Corporation <pharris[AT]opentext.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald[AT]wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

const char *bigreq_errors[] = {
  NULL
};

const char *composite_errors[] = {
  NULL
};

static const char *damage_errors[] = {
  "damage-BadDamage",
  NULL
};

const char *dpms_errors[] = {
  NULL
};

const char *dri2_errors[] = {
  NULL
};

const char *dri3_errors[] = {
  NULL
};

const char *ge_errors[] = {
  NULL
};

static const char *glx_errors[] = {
  "glx-BadContext",
  "glx-BadContextState",
  "glx-BadDrawable",
  "glx-BadPixmap",
  "glx-BadContextTag",
  "glx-BadCurrentWindow",
  "glx-BadRenderRequest",
  "glx-BadLargeRequest",
  "glx-UnsupportedPrivateRequest",
  "glx-BadFBConfig",
  "glx-BadPbuffer",
  "glx-BadCurrentDrawable",
  "glx-BadWindow",
  "glx-GLXBadProfileARB",
  NULL
};

const char *present_errors[] = {
  NULL
};

static const char *randr_errors[] = {
  "randr-BadOutput",
  "randr-BadCrtc",
  "randr-BadMode",
  "randr-BadProvider",
  NULL
};

static const char *record_errors[] = {
  "record-BadContext",
  NULL
};

static const char *render_errors[] = {
  "render-PictFormat",
  "render-Picture",
  "render-PictOp",
  "render-GlyphSet",
  "render-Glyph",
  NULL
};

const char *res_errors[] = {
  NULL
};

const char *screensaver_errors[] = {
  NULL
};

const char *shape_errors[] = {
  NULL
};

static const char *shm_errors[] = {
  "shm-BadSeg",
  NULL
};

static const char *sync_errors[] = {
  "sync-Counter",
  "sync-Alarm",
  NULL
};

const char *xc_misc_errors[] = {
  NULL
};

const char *xevie_errors[] = {
  NULL
};

const char *xf86dri_errors[] = {
  NULL
};

static const char *xf86vidmode_errors[] = {
  "xf86vidmode-BadClock",
  "xf86vidmode-BadHTimings",
  "xf86vidmode-BadVTimings",
  "xf86vidmode-ModeUnsuitable",
  "xf86vidmode-ExtensionDisabled",
  "xf86vidmode-ClientNotLocal",
  "xf86vidmode-ZoomLocked",
  NULL
};

static const char *xfixes_errors[] = {
  "xfixes-BadRegion",
  NULL
};

const char *xinerama_errors[] = {
  NULL
};

static const char *xinput_errors[] = {
  "xinput-Device",
  "xinput-Event",
  "xinput-Mode",
  "xinput-DeviceBusy",
  "xinput-Class",
  NULL
};

static const char *xkb_errors[] = {
  "xkb-Keyboard",
  NULL
};

static const char *xprint_errors[] = {
  "xprint-BadContext",
  "xprint-BadSequence",
  NULL
};

const char *xselinux_errors[] = {
  NULL
};

const char *xtest_errors[] = {
  NULL
};

static const char *xv_errors[] = {
  "xv-BadPort",
  "xv-BadEncoding",
  "xv-BadControl",
  NULL
};

const char *xvmc_errors[] = {
  NULL
};

