/* Do not modify this file. */
/* It was automatically generated by ../../tools/process-x11-xcb.pl
   using mesa version 20.1-branchpoint-2058-gf8110226baa */
/*
 * Copyright 2008, 2009, 2013, 2014 Open Text Corporation <pharris[AT]opentext.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald[AT]wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

static const value_string mesa_enum[] = {
  { 0x0000, "FALSE" },
  { 0x0001, "TRUE" },
  { 0x0002, "LINE_LOOP" },
  { 0x0003, "LINE_STRIP" },
  { 0x0004, "TRIANGLES" },
  { 0x0005, "TRIANGLE_STRIP" },
  { 0x0006, "TRIANGLE_FAN" },
  { 0x0007, "QUADS" },
  { 0x0008, "QUAD_STRIP" },
  { 0x0009, "POLYGON" },
  { 0x0010, "POLYGON_STIPPLE_BIT" },
  { 0x0020, "PIXEL_MODE_BIT" },
  { 0x0040, "LIGHTING_BIT" },
  { 0x0080, "FOG_BIT" },
  { 0x0100, "ACCUM" },
  { 0x0101, "LOAD" },
  { 0x0102, "RETURN" },
  { 0x0103, "MULT" },
  { 0x0104, "ADD" },
  { 0x0200, "NEVER" },
  { 0x0201, "LESS" },
  { 0x0202, "EQUAL" },
  { 0x0203, "LEQUAL" },
  { 0x0204, "GREATER" },
  { 0x0205, "NOTEQUAL" },
  { 0x0206, "GEQUAL" },
  { 0x0207, "ALWAYS" },
  { 0x0300, "SRC_COLOR" },
  { 0x0301, "ONE_MINUS_SRC_COLOR" },
  { 0x0302, "SRC_ALPHA" },
  { 0x0303, "ONE_MINUS_SRC_ALPHA" },
  { 0x0304, "DST_ALPHA" },
  { 0x0305, "ONE_MINUS_DST_ALPHA" },
  { 0x0306, "DST_COLOR" },
  { 0x0307, "ONE_MINUS_DST_COLOR" },
  { 0x0308, "SRC_ALPHA_SATURATE" },
  { 0x0400, "FRONT_LEFT" },
  { 0x0401, "FRONT_RIGHT" },
  { 0x0402, "BACK_LEFT" },
  { 0x0403, "BACK_RIGHT" },
  { 0x0404, "FRONT" },
  { 0x0405, "BACK" },
  { 0x0406, "LEFT" },
  { 0x0407, "RIGHT" },
  { 0x0408, "FRONT_AND_BACK" },
  { 0x0409, "AUX0" },
  { 0x040a, "AUX1" },
  { 0x040b, "AUX2" },
  { 0x040c, "AUX3" },
  { 0x0500, "INVALID_ENUM" },
  { 0x0501, "INVALID_VALUE" },
  { 0x0502, "INVALID_OPERATION" },
  { 0x0503, "STACK_OVERFLOW" },
  { 0x0504, "STACK_UNDERFLOW" },
  { 0x0505, "OUT_OF_MEMORY" },
  { 0x0600, "2D" },
  { 0x0601, "3D" },
  { 0x0602, "3D_COLOR" },
  { 0x0603, "3D_COLOR_TEXTURE" },
  { 0x0604, "4D_COLOR_TEXTURE" },
  { 0x0700, "PASS_THROUGH_TOKEN" },
  { 0x0701, "POINT_TOKEN" },
  { 0x0702, "LINE_TOKEN" },
  { 0x0703, "POLYGON_TOKEN" },
  { 0x0704, "BITMAP_TOKEN" },
  { 0x0705, "DRAW_PIXEL_TOKEN" },
  { 0x0706, "COPY_PIXEL_TOKEN" },
  { 0x0707, "LINE_RESET_TOKEN" },
  { 0x0800, "EXP" },
  { 0x0801, "EXP2" },
  { 0x0900, "CW" },
  { 0x0901, "CCW" },
  { 0x0a00, "COEFF" },
  { 0x0a01, "ORDER" },
  { 0x0a02, "DOMAIN" },
  { 0x0b00, "CURRENT_COLOR" },
  { 0x0b01, "CURRENT_INDEX" },
  { 0x0b02, "CURRENT_NORMAL" },
  { 0x0b03, "CURRENT_TEXTURE_COORDS" },
  { 0x0b04, "CURRENT_RASTER_COLOR" },
  { 0x0b05, "CURRENT_RASTER_INDEX" },
  { 0x0b06, "CURRENT_RASTER_TEXTURE_COORDS" },
  { 0x0b07, "CURRENT_RASTER_POSITION" },
  { 0x0b08, "CURRENT_RASTER_POSITION_VALID" },
  { 0x0b09, "CURRENT_RASTER_DISTANCE" },
  { 0x0b10, "POINT_SMOOTH" },
  { 0x0b11, "POINT_SIZE" },
  { 0x0b12, "POINT_SIZE_RANGE" },
  { 0x0b13, "POINT_SIZE_GRANULARITY" },
  { 0x0b20, "LINE_SMOOTH" },
  { 0x0b21, "LINE_WIDTH" },
  { 0x0b22, "LINE_WIDTH_RANGE" },
  { 0x0b23, "LINE_WIDTH_GRANULARITY" },
  { 0x0b24, "LINE_STIPPLE" },
  { 0x0b25, "LINE_STIPPLE_PATTERN" },
  { 0x0b26, "LINE_STIPPLE_REPEAT" },
  { 0x0b30, "LIST_MODE" },
  { 0x0b31, "MAX_LIST_NESTING" },
  { 0x0b32, "LIST_BASE" },
  { 0x0b33, "LIST_INDEX" },
  { 0x0b40, "POLYGON_MODE" },
  { 0x0b41, "POLYGON_SMOOTH" },
  { 0x0b42, "POLYGON_STIPPLE" },
  { 0x0b43, "EDGE_FLAG" },
  { 0x0b44, "CULL_FACE" },
  { 0x0b45, "CULL_FACE_MODE" },
  { 0x0b46, "FRONT_FACE" },
  { 0x0b50, "LIGHTING" },
  { 0x0b51, "LIGHT_MODEL_LOCAL_VIEWER" },
  { 0x0b52, "LIGHT_MODEL_TWO_SIDE" },
  { 0x0b53, "LIGHT_MODEL_AMBIENT" },
  { 0x0b54, "SHADE_MODEL" },
  { 0x0b55, "COLOR_MATERIAL_FACE" },
  { 0x0b56, "COLOR_MATERIAL_PARAMETER" },
  { 0x0b57, "COLOR_MATERIAL" },
  { 0x0b60, "FOG" },
  { 0x0b61, "FOG_INDEX" },
  { 0x0b62, "FOG_DENSITY" },
  { 0x0b63, "FOG_START" },
  { 0x0b64, "FOG_END" },
  { 0x0b65, "FOG_MODE" },
  { 0x0b66, "FOG_COLOR" },
  { 0x0b70, "DEPTH_RANGE" },
  { 0x0b71, "DEPTH_TEST" },
  { 0x0b72, "DEPTH_WRITEMASK" },
  { 0x0b73, "DEPTH_CLEAR_VALUE" },
  { 0x0b74, "DEPTH_FUNC" },
  { 0x0b80, "ACCUM_CLEAR_VALUE" },
  { 0x0b90, "STENCIL_TEST" },
  { 0x0b91, "STENCIL_CLEAR_VALUE" },
  { 0x0b92, "STENCIL_FUNC" },
  { 0x0b93, "STENCIL_VALUE_MASK" },
  { 0x0b94, "STENCIL_FAIL" },
  { 0x0b95, "STENCIL_PASS_DEPTH_FAIL" },
  { 0x0b96, "STENCIL_PASS_DEPTH_PASS" },
  { 0x0b97, "STENCIL_REF" },
  { 0x0b98, "STENCIL_WRITEMASK" },
  { 0x0ba0, "MATRIX_MODE" },
  { 0x0ba1, "NORMALIZE" },
  { 0x0ba2, "VIEWPORT" },
  { 0x0ba3, "MODELVIEW_STACK_DEPTH" },
  { 0x0ba4, "PROJECTION_STACK_DEPTH" },
  { 0x0ba5, "TEXTURE_STACK_DEPTH" },
  { 0x0ba6, "MODELVIEW_MATRIX" },
  { 0x0ba7, "PROJECTION_MATRIX" },
  { 0x0ba8, "TEXTURE_MATRIX" },
  { 0x0bb0, "ATTRIB_STACK_DEPTH" },
  { 0x0bb1, "CLIENT_ATTRIB_STACK_DEPTH" },
  { 0x0bc0, "ALPHA_TEST" },
  { 0x0bc1, "ALPHA_TEST_FUNC" },
  { 0x0bc2, "ALPHA_TEST_REF" },
  { 0x0bd0, "DITHER" },
  { 0x0be0, "BLEND_DST" },
  { 0x0be1, "BLEND_SRC" },
  { 0x0be2, "BLEND" },
  { 0x0bf0, "LOGIC_OP_MODE" },
  { 0x0bf1, "LOGIC_OP" },
  { 0x0bf2, "COLOR_LOGIC_OP" },
  { 0x0c00, "AUX_BUFFERS" },
  { 0x0c01, "DRAW_BUFFER" },
  { 0x0c02, "READ_BUFFER" },
  { 0x0c10, "SCISSOR_BOX" },
  { 0x0c11, "SCISSOR_TEST" },
  { 0x0c20, "INDEX_CLEAR_VALUE" },
  { 0x0c21, "INDEX_WRITEMASK" },
  { 0x0c22, "COLOR_CLEAR_VALUE" },
  { 0x0c23, "COLOR_WRITEMASK" },
  { 0x0c30, "INDEX_MODE" },
  { 0x0c31, "RGBA_MODE" },
  { 0x0c32, "DOUBLEBUFFER" },
  { 0x0c33, "STEREO" },
  { 0x0c40, "RENDER_MODE" },
  { 0x0c50, "PERSPECTIVE_CORRECTION_HINT" },
  { 0x0c51, "POINT_SMOOTH_HINT" },
  { 0x0c52, "LINE_SMOOTH_HINT" },
  { 0x0c53, "POLYGON_SMOOTH_HINT" },
  { 0x0c54, "FOG_HINT" },
  { 0x0c60, "TEXTURE_GEN_S" },
  { 0x0c61, "TEXTURE_GEN_T" },
  { 0x0c62, "TEXTURE_GEN_R" },
  { 0x0c63, "TEXTURE_GEN_Q" },
  { 0x0c70, "PIXEL_MAP_I_TO_I" },
  { 0x0c71, "PIXEL_MAP_S_TO_S" },
  { 0x0c72, "PIXEL_MAP_I_TO_R" },
  { 0x0c73, "PIXEL_MAP_I_TO_G" },
  { 0x0c74, "PIXEL_MAP_I_TO_B" },
  { 0x0c75, "PIXEL_MAP_I_TO_A" },
  { 0x0c76, "PIXEL_MAP_R_TO_R" },
  { 0x0c77, "PIXEL_MAP_G_TO_G" },
  { 0x0c78, "PIXEL_MAP_B_TO_B" },
  { 0x0c79, "PIXEL_MAP_A_TO_A" },
  { 0x0cb0, "PIXEL_MAP_I_TO_I_SIZE" },
  { 0x0cb1, "PIXEL_MAP_S_TO_S_SIZE" },
  { 0x0cb2, "PIXEL_MAP_I_TO_R_SIZE" },
  { 0x0cb3, "PIXEL_MAP_I_TO_G_SIZE" },
  { 0x0cb4, "PIXEL_MAP_I_TO_B_SIZE" },
  { 0x0cb5, "PIXEL_MAP_I_TO_A_SIZE" },
  { 0x0cb6, "PIXEL_MAP_R_TO_R_SIZE" },
  { 0x0cb7, "PIXEL_MAP_G_TO_G_SIZE" },
  { 0x0cb8, "PIXEL_MAP_B_TO_B_SIZE" },
  { 0x0cb9, "PIXEL_MAP_A_TO_A_SIZE" },
  { 0x0cf0, "UNPACK_SWAP_BYTES" },
  { 0x0cf1, "UNPACK_LSB_FIRST" },
  { 0x0cf2, "UNPACK_ROW_LENGTH" },
  { 0x0cf3, "UNPACK_SKIP_ROWS" },
  { 0x0cf4, "UNPACK_SKIP_PIXELS" },
  { 0x0cf5, "UNPACK_ALIGNMENT" },
  { 0x0d00, "PACK_SWAP_BYTES" },
  { 0x0d01, "PACK_LSB_FIRST" },
  { 0x0d02, "PACK_ROW_LENGTH" },
  { 0x0d03, "PACK_SKIP_ROWS" },
  { 0x0d04, "PACK_SKIP_PIXELS" },
  { 0x0d05, "PACK_ALIGNMENT" },
  { 0x0d10, "MAP_COLOR" },
  { 0x0d11, "MAP_STENCIL" },
  { 0x0d12, "INDEX_SHIFT" },
  { 0x0d13, "INDEX_OFFSET" },
  { 0x0d14, "RED_SCALE" },
  { 0x0d15, "RED_BIAS" },
  { 0x0d16, "ZOOM_X" },
  { 0x0d17, "ZOOM_Y" },
  { 0x0d18, "GREEN_SCALE" },
  { 0x0d19, "GREEN_BIAS" },
  { 0x0d1a, "BLUE_SCALE" },
  { 0x0d1b, "BLUE_BIAS" },
  { 0x0d1c, "ALPHA_SCALE" },
  { 0x0d1d, "ALPHA_BIAS" },
  { 0x0d1e, "DEPTH_SCALE" },
  { 0x0d1f, "DEPTH_BIAS" },
  { 0x0d30, "MAX_EVAL_ORDER" },
  { 0x0d31, "MAX_LIGHTS" },
  { 0x0d32, "MAX_CLIP_PLANES" },
  { 0x0d33, "MAX_TEXTURE_SIZE" },
  { 0x0d34, "MAX_PIXEL_MAP_TABLE" },
  { 0x0d35, "MAX_ATTRIB_STACK_DEPTH" },
  { 0x0d36, "MAX_MODELVIEW_STACK_DEPTH" },
  { 0x0d37, "MAX_NAME_STACK_DEPTH" },
  { 0x0d38, "MAX_PROJECTION_STACK_DEPTH" },
  { 0x0d39, "MAX_TEXTURE_STACK_DEPTH" },
  { 0x0d3a, "MAX_VIEWPORT_DIMS" },
  { 0x0d3b, "MAX_CLIENT_ATTRIB_STACK_DEPTH" },
  { 0x0d50, "SUBPIXEL_BITS" },
  { 0x0d51, "INDEX_BITS" },
  { 0x0d52, "RED_BITS" },
  { 0x0d53, "GREEN_BITS" },
  { 0x0d54, "BLUE_BITS" },
  { 0x0d55, "ALPHA_BITS" },
  { 0x0d56, "DEPTH_BITS" },
  { 0x0d57, "STENCIL_BITS" },
  { 0x0d58, "ACCUM_RED_BITS" },
  { 0x0d59, "ACCUM_GREEN_BITS" },
  { 0x0d5a, "ACCUM_BLUE_BITS" },
  { 0x0d5b, "ACCUM_ALPHA_BITS" },
  { 0x0d70, "NAME_STACK_DEPTH" },
  { 0x0d80, "AUTO_NORMAL" },
  { 0x0d90, "MAP1_COLOR_4" },
  { 0x0d91, "MAP1_INDEX" },
  { 0x0d92, "MAP1_NORMAL" },
  { 0x0d93, "MAP1_TEXTURE_COORD_1" },
  { 0x0d94, "MAP1_TEXTURE_COORD_2" },
  { 0x0d95, "MAP1_TEXTURE_COORD_3" },
  { 0x0d96, "MAP1_TEXTURE_COORD_4" },
  { 0x0d97, "MAP1_VERTEX_3" },
  { 0x0d98, "MAP1_VERTEX_4" },
  { 0x0db0, "MAP2_COLOR_4" },
  { 0x0db1, "MAP2_INDEX" },
  { 0x0db2, "MAP2_NORMAL" },
  { 0x0db3, "MAP2_TEXTURE_COORD_1" },
  { 0x0db4, "MAP2_TEXTURE_COORD_2" },
  { 0x0db5, "MAP2_TEXTURE_COORD_3" },
  { 0x0db6, "MAP2_TEXTURE_COORD_4" },
  { 0x0db7, "MAP2_VERTEX_3" },
  { 0x0db8, "MAP2_VERTEX_4" },
  { 0x0dd0, "MAP1_GRID_DOMAIN" },
  { 0x0dd1, "MAP1_GRID_SEGMENTS" },
  { 0x0dd2, "MAP2_GRID_DOMAIN" },
  { 0x0dd3, "MAP2_GRID_SEGMENTS" },
  { 0x0de0, "TEXTURE_1D" },
  { 0x0de1, "TEXTURE_2D" },
  { 0x0df0, "FEEDBACK_BUFFER_POINTER" },
  { 0x0df1, "FEEDBACK_BUFFER_SIZE" },
  { 0x0df2, "FEEDBACK_BUFFER_TYPE" },
  { 0x0df3, "SELECTION_BUFFER_POINTER" },
  { 0x0df4, "SELECTION_BUFFER_SIZE" },
  { 0x1000, "TEXTURE_WIDTH" },
  { 0x1001, "TEXTURE_HEIGHT" },
  { 0x1003, "TEXTURE_COMPONENTS" },
  { 0x1004, "TEXTURE_BORDER_COLOR" },
  { 0x1005, "TEXTURE_BORDER" },
  { 0x1100, "DONT_CARE" },
  { 0x1101, "FASTEST" },
  { 0x1102, "NICEST" },
  { 0x1200, "AMBIENT" },
  { 0x1201, "DIFFUSE" },
  { 0x1202, "SPECULAR" },
  { 0x1203, "POSITION" },
  { 0x1204, "SPOT_DIRECTION" },
  { 0x1205, "SPOT_EXPONENT" },
  { 0x1206, "SPOT_CUTOFF" },
  { 0x1207, "CONSTANT_ATTENUATION" },
  { 0x1208, "LINEAR_ATTENUATION" },
  { 0x1209, "QUADRATIC_ATTENUATION" },
  { 0x1300, "COMPILE" },
  { 0x1301, "COMPILE_AND_EXECUTE" },
  { 0x1400, "BYTE" },
  { 0x1401, "UNSIGNED_BYTE" },
  { 0x1402, "SHORT" },
  { 0x1403, "UNSIGNED_SHORT" },
  { 0x1404, "INT" },
  { 0x1405, "UNSIGNED_INT" },
  { 0x1406, "FLOAT" },
  { 0x1407, "2_BYTES" },
  { 0x1408, "3_BYTES" },
  { 0x1409, "4_BYTES" },
  { 0x140a, "DOUBLE" },
  { 0x140b, "HALF_FLOAT" },
  { 0x1500, "CLEAR" },
  { 0x1501, "AND" },
  { 0x1502, "AND_REVERSE" },
  { 0x1503, "COPY" },
  { 0x1504, "AND_INVERTED" },
  { 0x1505, "NOOP" },
  { 0x1506, "XOR" },
  { 0x1507, "OR" },
  { 0x1508, "NOR" },
  { 0x1509, "EQUIV" },
  { 0x150a, "INVERT" },
  { 0x150b, "OR_REVERSE" },
  { 0x150c, "COPY_INVERTED" },
  { 0x150d, "OR_INVERTED" },
  { 0x150e, "NAND" },
  { 0x150f, "SET" },
  { 0x1600, "EMISSION" },
  { 0x1601, "SHININESS" },
  { 0x1602, "AMBIENT_AND_DIFFUSE" },
  { 0x1603, "COLOR_INDEXES" },
  { 0x1700, "MODELVIEW" },
  { 0x1701, "PROJECTION" },
  { 0x1702, "TEXTURE" },
  { 0x1800, "COLOR" },
  { 0x1801, "DEPTH" },
  { 0x1802, "STENCIL" },
  { 0x1900, "COLOR_INDEX" },
  { 0x1901, "STENCIL_INDEX" },
  { 0x1902, "DEPTH_COMPONENT" },
  { 0x1903, "RED" },
  { 0x1904, "GREEN" },
  { 0x1905, "BLUE" },
  { 0x1906, "ALPHA" },
  { 0x1907, "RGB" },
  { 0x1908, "RGBA" },
  { 0x1909, "LUMINANCE" },
  { 0x190a, "LUMINANCE_ALPHA" },
  { 0x1a00, "BITMAP" },
  { 0x1b00, "POINT" },
  { 0x1b01, "LINE" },
  { 0x1b02, "FILL" },
  { 0x1c00, "RENDER" },
  { 0x1c01, "FEEDBACK" },
  { 0x1c02, "SELECT" },
  { 0x1d00, "FLAT" },
  { 0x1d01, "SMOOTH" },
  { 0x1e00, "KEEP" },
  { 0x1e01, "REPLACE" },
  { 0x1e02, "INCR" },
  { 0x1e03, "DECR" },
  { 0x1f00, "VENDOR" },
  { 0x1f01, "RENDERER" },
  { 0x1f02, "VERSION" },
  { 0x1f03, "EXTENSIONS" },
  { 0x2000, "S" },
  { 0x2001, "T" },
  { 0x2002, "R" },
  { 0x2003, "Q" },
  { 0x2100, "MODULATE" },
  { 0x2101, "DECAL" },
  { 0x2200, "TEXTURE_ENV_MODE" },
  { 0x2201, "TEXTURE_ENV_COLOR" },
  { 0x2300, "TEXTURE_ENV" },
  { 0x2400, "EYE_LINEAR" },
  { 0x2401, "OBJECT_LINEAR" },
  { 0x2402, "SPHERE_MAP" },
  { 0x2500, "TEXTURE_GEN_MODE" },
  { 0x2501, "OBJECT_PLANE" },
  { 0x2502, "EYE_PLANE" },
  { 0x2600, "NEAREST" },
  { 0x2601, "LINEAR" },
  { 0x2700, "NEAREST_MIPMAP_NEAREST" },
  { 0x2701, "LINEAR_MIPMAP_NEAREST" },
  { 0x2702, "NEAREST_MIPMAP_LINEAR" },
  { 0x2703, "LINEAR_MIPMAP_LINEAR" },
  { 0x2800, "TEXTURE_MAG_FILTER" },
  { 0x2801, "TEXTURE_MIN_FILTER" },
  { 0x2802, "TEXTURE_WRAP_S" },
  { 0x2803, "TEXTURE_WRAP_T" },
  { 0x2900, "CLAMP" },
  { 0x2901, "REPEAT" },
  { 0x2a00, "POLYGON_OFFSET_UNITS" },
  { 0x2a01, "POLYGON_OFFSET_POINT" },
  { 0x2a02, "POLYGON_OFFSET_LINE" },
  { 0x2a10, "R3_G3_B2" },
  { 0x2a20, "V2F" },
  { 0x2a21, "V3F" },
  { 0x2a22, "C4UB_V2F" },
  { 0x2a23, "C4UB_V3F" },
  { 0x2a24, "C3F_V3F" },
  { 0x2a25, "N3F_V3F" },
  { 0x2a26, "C4F_N3F_V3F" },
  { 0x2a27, "T2F_V3F" },
  { 0x2a28, "T4F_V4F" },
  { 0x2a29, "T2F_C4UB_V3F" },
  { 0x2a2a, "T2F_C3F_V3F" },
  { 0x2a2b, "T2F_N3F_V3F" },
  { 0x2a2c, "T2F_C4F_N3F_V3F" },
  { 0x2a2d, "T4F_C4F_N3F_V4F" },
  { 0x3000, "CLIP_PLANE0" },
  { 0x3001, "CLIP_PLANE1" },
  { 0x3002, "CLIP_PLANE2" },
  { 0x3003, "CLIP_PLANE3" },
  { 0x3004, "CLIP_PLANE4" },
  { 0x3005, "CLIP_PLANE5" },
  { 0x4000, "LIGHT0" },
  { 0x4001, "LIGHT1" },
  { 0x4002, "LIGHT2" },
  { 0x4003, "LIGHT3" },
  { 0x4004, "LIGHT4" },
  { 0x4005, "LIGHT5" },
  { 0x4006, "LIGHT6" },
  { 0x4007, "LIGHT7" },
  { 0x8000, "HINT_BIT" },
  { 0x8001, "CONSTANT_COLOR" },
  { 0x8002, "ONE_MINUS_CONSTANT_COLOR" },
  { 0x8003, "CONSTANT_ALPHA" },
  { 0x8004, "ONE_MINUS_CONSTANT_ALPHA" },
  { 0x8005, "BLEND_COLOR" },
  { 0x8006, "FUNC_ADD" },
  { 0x8007, "MIN" },
  { 0x8008, "MAX" },
  { 0x8009, "BLEND_EQUATION" },
  { 0x800a, "FUNC_SUBTRACT" },
  { 0x800b, "FUNC_REVERSE_SUBTRACT" },
  { 0x8010, "CONVOLUTION_1D" },
  { 0x8011, "CONVOLUTION_2D" },
  { 0x8012, "SEPARABLE_2D" },
  { 0x8013, "CONVOLUTION_BORDER_MODE" },
  { 0x8014, "CONVOLUTION_FILTER_SCALE" },
  { 0x8015, "CONVOLUTION_FILTER_BIAS" },
  { 0x8016, "REDUCE" },
  { 0x8017, "CONVOLUTION_FORMAT" },
  { 0x8018, "CONVOLUTION_WIDTH" },
  { 0x8019, "CONVOLUTION_HEIGHT" },
  { 0x801a, "MAX_CONVOLUTION_WIDTH" },
  { 0x801b, "MAX_CONVOLUTION_HEIGHT" },
  { 0x801c, "POST_CONVOLUTION_RED_SCALE" },
  { 0x801d, "POST_CONVOLUTION_GREEN_SCALE" },
  { 0x801e, "POST_CONVOLUTION_BLUE_SCALE" },
  { 0x801f, "POST_CONVOLUTION_ALPHA_SCALE" },
  { 0x8020, "POST_CONVOLUTION_RED_BIAS" },
  { 0x8021, "POST_CONVOLUTION_GREEN_BIAS" },
  { 0x8022, "POST_CONVOLUTION_BLUE_BIAS" },
  { 0x8023, "POST_CONVOLUTION_ALPHA_BIAS" },
  { 0x8024, "HISTOGRAM" },
  { 0x8025, "PROXY_HISTOGRAM" },
  { 0x8026, "HISTOGRAM_WIDTH" },
  { 0x8027, "HISTOGRAM_FORMAT" },
  { 0x8028, "HISTOGRAM_RED_SIZE" },
  { 0x8029, "HISTOGRAM_GREEN_SIZE" },
  { 0x802a, "HISTOGRAM_BLUE_SIZE" },
  { 0x802b, "HISTOGRAM_ALPHA_SIZE" },
  { 0x802c, "HISTOGRAM_LUMINANCE_SIZE" },
  { 0x802d, "HISTOGRAM_SINK" },
  { 0x802e, "MINMAX" },
  { 0x802f, "MINMAX_FORMAT" },
  { 0x8030, "MINMAX_SINK" },
  { 0x8031, "TABLE_TOO_LARGE_EXT" },
  { 0x8032, "UNSIGNED_BYTE_3_3_2" },
  { 0x8033, "UNSIGNED_SHORT_4_4_4_4" },
  { 0x8034, "UNSIGNED_SHORT_5_5_5_1" },
  { 0x8035, "UNSIGNED_INT_8_8_8_8" },
  { 0x8036, "UNSIGNED_INT_10_10_10_2" },
  { 0x8037, "POLYGON_OFFSET_FILL" },
  { 0x8038, "POLYGON_OFFSET_FACTOR" },
  { 0x803a, "RESCALE_NORMAL" },
  { 0x803b, "ALPHA4" },
  { 0x803c, "ALPHA8" },
  { 0x803d, "ALPHA12" },
  { 0x803e, "ALPHA16" },
  { 0x803f, "LUMINANCE4" },
  { 0x8040, "LUMINANCE8" },
  { 0x8041, "LUMINANCE12" },
  { 0x8042, "LUMINANCE16" },
  { 0x8043, "LUMINANCE4_ALPHA4" },
  { 0x8044, "LUMINANCE6_ALPHA2" },
  { 0x8045, "LUMINANCE8_ALPHA8" },
  { 0x8046, "LUMINANCE12_ALPHA4" },
  { 0x8047, "LUMINANCE12_ALPHA12" },
  { 0x8048, "LUMINANCE16_ALPHA16" },
  { 0x8049, "INTENSITY" },
  { 0x804a, "INTENSITY4" },
  { 0x804b, "INTENSITY8" },
  { 0x804c, "INTENSITY12" },
  { 0x804d, "INTENSITY16" },
  { 0x804e, "RGB2_EXT" },
  { 0x804f, "RGB4" },
  { 0x8050, "RGB5" },
  { 0x8051, "RGB8" },
  { 0x8052, "RGB10" },
  { 0x8053, "RGB12" },
  { 0x8054, "RGB16" },
  { 0x8055, "RGBA2" },
  { 0x8056, "RGBA4" },
  { 0x8057, "RGB5_A1" },
  { 0x8058, "RGBA8" },
  { 0x8059, "RGB10_A2" },
  { 0x805a, "RGBA12" },
  { 0x805b, "RGBA16" },
  { 0x805c, "TEXTURE_RED_SIZE" },
  { 0x805d, "TEXTURE_GREEN_SIZE" },
  { 0x805e, "TEXTURE_BLUE_SIZE" },
  { 0x805f, "TEXTURE_ALPHA_SIZE" },
  { 0x8060, "TEXTURE_LUMINANCE_SIZE" },
  { 0x8061, "TEXTURE_INTENSITY_SIZE" },
  { 0x8062, "REPLACE_EXT" },
  { 0x8063, "PROXY_TEXTURE_1D" },
  { 0x8064, "PROXY_TEXTURE_2D" },
  { 0x8065, "TEXTURE_TOO_LARGE_EXT" },
  { 0x8066, "TEXTURE_PRIORITY" },
  { 0x8067, "TEXTURE_RESIDENT" },
  { 0x8068, "TEXTURE_BINDING_1D" },
  { 0x8069, "TEXTURE_BINDING_2D" },
  { 0x806a, "TEXTURE_BINDING_3D" },
  { 0x806b, "PACK_SKIP_IMAGES" },
  { 0x806c, "PACK_IMAGE_HEIGHT" },
  { 0x806d, "UNPACK_SKIP_IMAGES" },
  { 0x806e, "UNPACK_IMAGE_HEIGHT" },
  { 0x806f, "TEXTURE_3D" },
  { 0x8070, "PROXY_TEXTURE_3D" },
  { 0x8071, "TEXTURE_DEPTH" },
  { 0x8072, "TEXTURE_WRAP_R" },
  { 0x8073, "MAX_3D_TEXTURE_SIZE" },
  { 0x8074, "VERTEX_ARRAY" },
  { 0x8075, "NORMAL_ARRAY" },
  { 0x8076, "COLOR_ARRAY" },
  { 0x8077, "INDEX_ARRAY" },
  { 0x8078, "TEXTURE_COORD_ARRAY" },
  { 0x8079, "EDGE_FLAG_ARRAY" },
  { 0x807a, "VERTEX_ARRAY_SIZE" },
  { 0x807b, "VERTEX_ARRAY_TYPE" },
  { 0x807c, "VERTEX_ARRAY_STRIDE" },
  { 0x807d, "VERTEX_ARRAY_COUNT_EXT" },
  { 0x807e, "NORMAL_ARRAY_TYPE" },
  { 0x807f, "NORMAL_ARRAY_STRIDE" },
  { 0x8080, "NORMAL_ARRAY_COUNT_EXT" },
  { 0x8081, "COLOR_ARRAY_SIZE" },
  { 0x8082, "COLOR_ARRAY_TYPE" },
  { 0x8083, "COLOR_ARRAY_STRIDE" },
  { 0x8084, "COLOR_ARRAY_COUNT_EXT" },
  { 0x8085, "INDEX_ARRAY_TYPE" },
  { 0x8086, "INDEX_ARRAY_STRIDE" },
  { 0x8087, "INDEX_ARRAY_COUNT_EXT" },
  { 0x8088, "TEXTURE_COORD_ARRAY_SIZE" },
  { 0x8089, "TEXTURE_COORD_ARRAY_TYPE" },
  { 0x808a, "TEXTURE_COORD_ARRAY_STRIDE" },
  { 0x808b, "TEXTURE_COORD_ARRAY_COUNT_EXT" },
  { 0x808c, "EDGE_FLAG_ARRAY_STRIDE" },
  { 0x808d, "EDGE_FLAG_ARRAY_COUNT_EXT" },
  { 0x808e, "VERTEX_ARRAY_POINTER" },
  { 0x808f, "NORMAL_ARRAY_POINTER" },
  { 0x8090, "COLOR_ARRAY_POINTER" },
  { 0x8091, "INDEX_ARRAY_POINTER" },
  { 0x8092, "TEXTURE_COORD_ARRAY_POINTER" },
  { 0x8093, "EDGE_FLAG_ARRAY_POINTER" },
  { 0x809d, "MULTISAMPLE" },
  { 0x809e, "SAMPLE_ALPHA_TO_COVERAGE" },
  { 0x809f, "SAMPLE_ALPHA_TO_ONE" },
  { 0x80a0, "SAMPLE_COVERAGE" },
  { 0x80a8, "SAMPLE_BUFFERS" },
  { 0x80a9, "SAMPLES" },
  { 0x80aa, "SAMPLE_COVERAGE_VALUE" },
  { 0x80ab, "SAMPLE_COVERAGE_INVERT" },
  { 0x80b1, "COLOR_MATRIX" },
  { 0x80b2, "COLOR_MATRIX_STACK_DEPTH" },
  { 0x80b3, "MAX_COLOR_MATRIX_STACK_DEPTH" },
  { 0x80b4, "POST_COLOR_MATRIX_RED_SCALE" },
  { 0x80b5, "POST_COLOR_MATRIX_GREEN_SCALE" },
  { 0x80b6, "POST_COLOR_MATRIX_BLUE_SCALE" },
  { 0x80b7, "POST_COLOR_MATRIX_ALPHA_SCALE" },
  { 0x80b8, "POST_COLOR_MATRIX_RED_BIAS" },
  { 0x80b9, "POST_COLOR_MATRIX_GREEN_BIAS" },
  { 0x80ba, "POST_COLOR_MATRIX_BLUE_BIAS" },
  { 0x80bb, "POST_COLOR_MATRIX_ALPHA_BIAS" },
  { 0x80bc, "TEXTURE_COLOR_TABLE_SGI" },
  { 0x80bd, "PROXY_TEXTURE_COLOR_TABLE_SGI" },
  { 0x80bf, "TEXTURE_COMPARE_FAIL_VALUE_ARB" },
  { 0x80c8, "BLEND_DST_RGB" },
  { 0x80c9, "BLEND_SRC_RGB" },
  { 0x80ca, "BLEND_DST_ALPHA" },
  { 0x80cb, "BLEND_SRC_ALPHA" },
  { 0x80d0, "COLOR_TABLE" },
  { 0x80d1, "POST_CONVOLUTION_COLOR_TABLE" },
  { 0x80d2, "POST_COLOR_MATRIX_COLOR_TABLE" },
  { 0x80d3, "PROXY_COLOR_TABLE" },
  { 0x80d4, "PROXY_POST_CONVOLUTION_COLOR_TABLE" },
  { 0x80d5, "PROXY_POST_COLOR_MATRIX_COLOR_TABLE" },
  { 0x80d6, "COLOR_TABLE_SCALE" },
  { 0x80d7, "COLOR_TABLE_BIAS" },
  { 0x80d8, "COLOR_TABLE_FORMAT" },
  { 0x80d9, "COLOR_TABLE_WIDTH" },
  { 0x80da, "COLOR_TABLE_RED_SIZE" },
  { 0x80db, "COLOR_TABLE_GREEN_SIZE" },
  { 0x80dc, "COLOR_TABLE_BLUE_SIZE" },
  { 0x80dd, "COLOR_TABLE_ALPHA_SIZE" },
  { 0x80de, "COLOR_TABLE_LUMINANCE_SIZE" },
  { 0x80df, "COLOR_TABLE_INTENSITY_SIZE" },
  { 0x80e0, "BGR" },
  { 0x80e1, "BGRA" },
  { 0x80e8, "MAX_ELEMENTS_VERTICES" },
  { 0x80e9, "MAX_ELEMENTS_INDICES" },
  { 0x80ed, "TEXTURE_INDEX_SIZE_EXT" },
  { 0x80f0, "CLIP_VOLUME_CLIPPING_HINT_EXT" },
  { 0x8126, "POINT_SIZE_MIN" },
  { 0x8127, "POINT_SIZE_MAX" },
  { 0x8128, "POINT_FADE_THRESHOLD_SIZE" },
  { 0x8129, "POINT_DISTANCE_ATTENUATION" },
  { 0x812d, "CLAMP_TO_BORDER" },
  { 0x812f, "CLAMP_TO_EDGE" },
  { 0x813a, "TEXTURE_MIN_LOD" },
  { 0x813b, "TEXTURE_MAX_LOD" },
  { 0x813c, "TEXTURE_BASE_LEVEL" },
  { 0x813d, "TEXTURE_MAX_LEVEL" },
  { 0x8150, "IGNORE_BORDER_HP" },
  { 0x8151, "CONSTANT_BORDER_HP" },
  { 0x8153, "REPLICATE_BORDER_HP" },
  { 0x8154, "CONVOLUTION_BORDER_COLOR" },
  { 0x8165, "OCCLUSION_TEST_HP" },
  { 0x8166, "OCCLUSION_TEST_RESULT_HP" },
  { 0x8170, "LINEAR_CLIPMAP_LINEAR_SGIX" },
  { 0x8171, "TEXTURE_CLIPMAP_CENTER_SGIX" },
  { 0x8172, "TEXTURE_CLIPMAP_FRAME_SGIX" },
  { 0x8173, "TEXTURE_CLIPMAP_OFFSET_SGIX" },
  { 0x8174, "TEXTURE_CLIPMAP_VIRTUAL_DEPTH_SGIX" },
  { 0x8175, "TEXTURE_CLIPMAP_LOD_OFFSET_SGIX" },
  { 0x8176, "TEXTURE_CLIPMAP_DEPTH_SGIX" },
  { 0x8177, "MAX_CLIPMAP_DEPTH_SGIX" },
  { 0x8178, "MAX_CLIPMAP_VIRTUAL_DEPTH_SGIX" },
  { 0x8179, "POST_TEXTURE_FILTER_BIAS_SGIX" },
  { 0x817a, "POST_TEXTURE_FILTER_SCALE_SGIX" },
  { 0x817b, "POST_TEXTURE_FILTER_BIAS_RANGE_SGIX" },
  { 0x817c, "POST_TEXTURE_FILTER_SCALE_RANGE_SGIX" },
  { 0x818e, "TEXTURE_LOD_BIAS_S_SGIX" },
  { 0x818f, "TEXTURE_LOD_BIAS_T_SGIX" },
  { 0x8190, "TEXTURE_LOD_BIAS_R_SGIX" },
  { 0x8191, "GENERATE_MIPMAP" },
  { 0x8192, "GENERATE_MIPMAP_HINT" },
  { 0x8198, "FOG_OFFSET_SGIX" },
  { 0x8199, "FOG_OFFSET_VALUE_SGIX" },
  { 0x819a, "TEXTURE_COMPARE_SGIX" },
  { 0x819b, "TEXTURE_COMPARE_OPERATOR_SGIX" },
  { 0x819c, "TEXTURE_LEQUAL_R_SGIX" },
  { 0x819d, "TEXTURE_GEQUAL_R_SGIX" },
  { 0x81a5, "DEPTH_COMPONENT16" },
  { 0x81a6, "DEPTH_COMPONENT24" },
  { 0x81a7, "DEPTH_COMPONENT32" },
  { 0x81a8, "ARRAY_ELEMENT_LOCK_FIRST_EXT" },
  { 0x81a9, "ARRAY_ELEMENT_LOCK_COUNT_EXT" },
  { 0x81aa, "CULL_VERTEX_EXT" },
  { 0x81ab, "CULL_VERTEX_EYE_POSITION_EXT" },
  { 0x81ac, "CULL_VERTEX_OBJECT_POSITION_EXT" },
  { 0x81d4, "WRAP_BORDER_SUN" },
  { 0x81ef, "TEXTURE_COLOR_WRITEMASK_SGIS" },
  { 0x81f8, "LIGHT_MODEL_COLOR_CONTROL" },
  { 0x81f9, "SINGLE_COLOR" },
  { 0x81fa, "SEPARATE_SPECULAR_COLOR" },
  { 0x81fb, "SHARED_TEXTURE_PALETTE_EXT" },
  { 0x821f, "BUFFER_IMMUTABLE_STORAGE" },
  { 0x8220, "BUFFER_STORAGE_FLAGS" },
  { 0x826e, "MAX_UNIFORM_LOCATIONS" },
  { 0x82f9, "MAX_CULL_DISTANCES" },
  { 0x82fa, "MAX_COMBINED_CLIP_AND_CULL_DISTANCES" },
  { 0x8362, "UNSIGNED_BYTE_2_3_3_REV" },
  { 0x8363, "UNSIGNED_SHORT_5_6_5" },
  { 0x8364, "UNSIGNED_SHORT_5_6_5_REV" },
  { 0x8365, "UNSIGNED_SHORT_4_4_4_4_REV" },
  { 0x8366, "UNSIGNED_SHORT_1_5_5_5_REV" },
  { 0x8367, "UNSIGNED_INT_8_8_8_8_REV" },
  { 0x8368, "UNSIGNED_INT_2_10_10_10_REV" },
  { 0x8369, "TEXTURE_MAX_CLAMP_S_SGIX" },
  { 0x836a, "TEXTURE_MAX_CLAMP_T_SGIX" },
  { 0x836b, "TEXTURE_MAX_CLAMP_R_SGIX" },
  { 0x8370, "MIRRORED_REPEAT" },
  { 0x83a0, "RGB_S3TC" },
  { 0x83a1, "RGB4_S3TC" },
  { 0x83a2, "RGBA_S3TC" },
  { 0x83a3, "RGBA4_S3TC" },
  { 0x83a4, "RGBA_DXT5_S3TC" },
  { 0x83a5, "RGBA4_DXT5_S3TC" },
  { 0x83f0, "COMPRESSED_RGB_S3TC_DXT1_EXT" },
  { 0x83f1, "COMPRESSED_RGBA_S3TC_DXT1_EXT" },
  { 0x83f2, "COMPRESSED_RGBA_S3TC_DXT3_EXT" },
  { 0x83f3, "COMPRESSED_RGBA_S3TC_DXT5_EXT" },
  { 0x83fe, "CONSERVATIVE_RASTERIZATION_INTEL" },
  { 0x844d, "NEAREST_CLIPMAP_NEAREST_SGIX" },
  { 0x844e, "NEAREST_CLIPMAP_LINEAR_SGIX" },
  { 0x844f, "LINEAR_CLIPMAP_NEAREST_SGIX" },
  { 0x8450, "FOG_COORDINATE_SOURCE" },
  { 0x8451, "FOG_COORDINATE" },
  { 0x8452, "FRAGMENT_DEPTH" },
  { 0x8453, "CURRENT_FOG_COORDINATE" },
  { 0x8454, "FOG_COORDINATE_ARRAY_TYPE" },
  { 0x8455, "FOG_COORDINATE_ARRAY_STRIDE" },
  { 0x8456, "FOG_COORDINATE_ARRAY_POINTER" },
  { 0x8457, "FOG_COORDINATE_ARRAY" },
  { 0x8458, "COLOR_SUM" },
  { 0x8459, "CURRENT_SECONDARY_COLOR" },
  { 0x845a, "SECONDARY_COLOR_ARRAY_SIZE" },
  { 0x845b, "SECONDARY_COLOR_ARRAY_TYPE" },
  { 0x845c, "SECONDARY_COLOR_ARRAY_STRIDE" },
  { 0x845d, "SECONDARY_COLOR_ARRAY_POINTER" },
  { 0x845e, "SECONDARY_COLOR_ARRAY" },
  { 0x845f, "CURRENT_RASTER_SECONDARY_COLOR" },
  { 0x846d, "ALIASED_POINT_SIZE_RANGE" },
  { 0x846e, "ALIASED_LINE_WIDTH_RANGE" },
  { 0x84c0, "TEXTURE0" },
  { 0x84c1, "TEXTURE1" },
  { 0x84c2, "TEXTURE2" },
  { 0x84c3, "TEXTURE3" },
  { 0x84c4, "TEXTURE4" },
  { 0x84c5, "TEXTURE5" },
  { 0x84c6, "TEXTURE6" },
  { 0x84c7, "TEXTURE7" },
  { 0x84c8, "TEXTURE8" },
  { 0x84c9, "TEXTURE9" },
  { 0x84ca, "TEXTURE10" },
  { 0x84cb, "TEXTURE11" },
  { 0x84cc, "TEXTURE12" },
  { 0x84cd, "TEXTURE13" },
  { 0x84ce, "TEXTURE14" },
  { 0x84cf, "TEXTURE15" },
  { 0x84d0, "TEXTURE16" },
  { 0x84d1, "TEXTURE17" },
  { 0x84d2, "TEXTURE18" },
  { 0x84d3, "TEXTURE19" },
  { 0x84d4, "TEXTURE20" },
  { 0x84d5, "TEXTURE21" },
  { 0x84d6, "TEXTURE22" },
  { 0x84d7, "TEXTURE23" },
  { 0x84d8, "TEXTURE24" },
  { 0x84d9, "TEXTURE25" },
  { 0x84da, "TEXTURE26" },
  { 0x84db, "TEXTURE27" },
  { 0x84dc, "TEXTURE28" },
  { 0x84dd, "TEXTURE29" },
  { 0x84de, "TEXTURE30" },
  { 0x84df, "TEXTURE31" },
  { 0x84e0, "ACTIVE_TEXTURE" },
  { 0x84e1, "CLIENT_ACTIVE_TEXTURE" },
  { 0x84e2, "MAX_TEXTURE_UNITS" },
  { 0x84e3, "TRANSPOSE_MODELVIEW_MATRIX" },
  { 0x84e4, "TRANSPOSE_PROJECTION_MATRIX" },
  { 0x84e5, "TRANSPOSE_TEXTURE_MATRIX" },
  { 0x84e6, "TRANSPOSE_COLOR_MATRIX" },
  { 0x84e7, "SUBTRACT" },
  { 0x84e9, "COMPRESSED_ALPHA" },
  { 0x84ea, "COMPRESSED_LUMINANCE" },
  { 0x84eb, "COMPRESSED_LUMINANCE_ALPHA" },
  { 0x84ec, "COMPRESSED_INTENSITY" },
  { 0x84ed, "COMPRESSED_RGB" },
  { 0x84ee, "COMPRESSED_RGBA" },
  { 0x84ef, "TEXTURE_COMPRESSION_HINT" },
  { 0x84f5, "TEXTURE_RECTANGLE_ARB" },
  { 0x84f6, "TEXTURE_BINDING_RECTANGLE_ARB" },
  { 0x84f7, "PROXY_TEXTURE_RECTANGLE_ARB" },
  { 0x84f8, "MAX_RECTANGLE_TEXTURE_SIZE_ARB" },
  { 0x84f9, "DEPTH_STENCIL_NV" },
  { 0x84fa, "UNSIGNED_INT_24_8_NV" },
  { 0x84fd, "MAX_TEXTURE_LOD_BIAS" },
  { 0x84fe, "TEXTURE_MAX_ANISOTROPY_EXT" },
  { 0x84ff, "MAX_TEXTURE_MAX_ANISOTROPY_EXT" },
  { 0x8500, "TEXTURE_FILTER_CONTROL" },
  { 0x8501, "TEXTURE_LOD_BIAS" },
  { 0x8503, "COMBINE4_NV" },
  { 0x8504, "MAX_SHININESS_NV" },
  { 0x8505, "MAX_SPOT_EXPONENT_NV" },
  { 0x8507, "INCR_WRAP" },
  { 0x8508, "DECR_WRAP" },
  { 0x850a, "MODELVIEW1_ARB" },
  { 0x8511, "NORMAL_MAP" },
  { 0x8512, "REFLECTION_MAP" },
  { 0x8513, "TEXTURE_CUBE_MAP" },
  { 0x8514, "TEXTURE_BINDING_CUBE_MAP" },
  { 0x8515, "TEXTURE_CUBE_MAP_POSITIVE_X" },
  { 0x8516, "TEXTURE_CUBE_MAP_NEGATIVE_X" },
  { 0x8517, "TEXTURE_CUBE_MAP_POSITIVE_Y" },
  { 0x8518, "TEXTURE_CUBE_MAP_NEGATIVE_Y" },
  { 0x8519, "TEXTURE_CUBE_MAP_POSITIVE_Z" },
  { 0x851a, "TEXTURE_CUBE_MAP_NEGATIVE_Z" },
  { 0x851b, "PROXY_TEXTURE_CUBE_MAP" },
  { 0x851c, "MAX_CUBE_MAP_TEXTURE_SIZE" },
  { 0x8534, "MULTISAMPLE_FILTER_HINT_NV" },
  { 0x855a, "FOG_DISTANCE_MODE_NV" },
  { 0x855b, "EYE_RADIAL_NV" },
  { 0x855c, "EYE_PLANE_ABSOLUTE_NV" },
  { 0x8570, "COMBINE" },
  { 0x8571, "COMBINE_RGB" },
  { 0x8572, "COMBINE_ALPHA" },
  { 0x8573, "RGB_SCALE" },
  { 0x8574, "ADD_SIGNED" },
  { 0x8575, "INTERPOLATE" },
  { 0x8576, "CONSTANT" },
  { 0x8577, "PRIMARY_COLOR" },
  { 0x8578, "PREVIOUS" },
  { 0x8580, "SOURCE0_RGB" },
  { 0x8581, "SOURCE1_RGB" },
  { 0x8582, "SOURCE2_RGB" },
  { 0x8583, "SOURCE3_RGB_NV" },
  { 0x8588, "SOURCE0_ALPHA" },
  { 0x8589, "SOURCE1_ALPHA" },
  { 0x858a, "SOURCE2_ALPHA" },
  { 0x858b, "SOURCE3_ALPHA_NV" },
  { 0x8590, "OPERAND0_RGB" },
  { 0x8591, "OPERAND1_RGB" },
  { 0x8592, "OPERAND2_RGB" },
  { 0x8593, "OPERAND3_RGB_NV" },
  { 0x8598, "OPERAND0_ALPHA" },
  { 0x8599, "OPERAND1_ALPHA" },
  { 0x859a, "OPERAND2_ALPHA" },
  { 0x859b, "OPERAND3_ALPHA_NV" },
  { 0x85b7, "TEXTURE_RANGE_LENGTH_APPLE" },
  { 0x85b8, "TEXTURE_RANGE_POINTER_APPLE" },
  { 0x85b9, "YCBCR_422_APPLE" },
  { 0x85ba, "UNSIGNED_SHORT_8_8_APPLE" },
  { 0x85bb, "UNSIGNED_SHORT_8_8_REV_APPLE" },
  { 0x85bc, "TEXTURE_STORAGE_HINT_APPLE" },
  { 0x85bd, "STORAGE_PRIVATE_APPLE" },
  { 0x85be, "STORAGE_CACHED_APPLE" },
  { 0x85bf, "STORAGE_SHARED_APPLE" },
  { 0x85cc, "SLICE_ACCUM_SUN" },
  { 0x8614, "QUAD_MESH_SUN" },
  { 0x8615, "TRIANGLE_MESH_SUN" },
  { 0x8620, "VERTEX_PROGRAM_ARB" },
  { 0x8621, "VERTEX_STATE_PROGRAM_NV" },
  { 0x8622, "VERTEX_ATTRIB_ARRAY_ENABLED" },
  { 0x8623, "VERTEX_ATTRIB_ARRAY_SIZE" },
  { 0x8624, "VERTEX_ATTRIB_ARRAY_STRIDE" },
  { 0x8625, "VERTEX_ATTRIB_ARRAY_TYPE" },
  { 0x8626, "CURRENT_VERTEX_ATTRIB" },
  { 0x8627, "PROGRAM_LENGTH_ARB" },
  { 0x8628, "PROGRAM_STRING_ARB" },
  { 0x8629, "MODELVIEW_PROJECTION_NV" },
  { 0x862a, "IDENTITY_NV" },
  { 0x862b, "INVERSE_NV" },
  { 0x862c, "TRANSPOSE_NV" },
  { 0x862d, "INVERSE_TRANSPOSE_NV" },
  { 0x862e, "MAX_PROGRAM_MATRIX_STACK_DEPTH_ARB" },
  { 0x862f, "MAX_PROGRAM_MATRICES_ARB" },
  { 0x8630, "MATRIX0_NV" },
  { 0x8631, "MATRIX1_NV" },
  { 0x8632, "MATRIX2_NV" },
  { 0x8633, "MATRIX3_NV" },
  { 0x8634, "MATRIX4_NV" },
  { 0x8635, "MATRIX5_NV" },
  { 0x8636, "MATRIX6_NV" },
  { 0x8637, "MATRIX7_NV" },
  { 0x8640, "CURRENT_MATRIX_STACK_DEPTH_ARB" },
  { 0x8641, "CURRENT_MATRIX_ARB" },
  { 0x8642, "VERTEX_PROGRAM_POINT_SIZE" },
  { 0x8643, "VERTEX_PROGRAM_TWO_SIDE" },
  { 0x8644, "PROGRAM_PARAMETER_NV" },
  { 0x8645, "VERTEX_ATTRIB_ARRAY_POINTER" },
  { 0x8646, "PROGRAM_TARGET_NV" },
  { 0x8647, "PROGRAM_RESIDENT_NV" },
  { 0x8648, "TRACK_MATRIX_NV" },
  { 0x8649, "TRACK_MATRIX_TRANSFORM_NV" },
  { 0x864a, "VERTEX_PROGRAM_BINDING_NV" },
  { 0x864b, "PROGRAM_ERROR_POSITION_ARB" },
  { 0x864f, "DEPTH_CLAMP_NV" },
  { 0x8650, "VERTEX_ATTRIB_ARRAY0_NV" },
  { 0x8651, "VERTEX_ATTRIB_ARRAY1_NV" },
  { 0x8652, "VERTEX_ATTRIB_ARRAY2_NV" },
  { 0x8653, "VERTEX_ATTRIB_ARRAY3_NV" },
  { 0x8654, "VERTEX_ATTRIB_ARRAY4_NV" },
  { 0x8655, "VERTEX_ATTRIB_ARRAY5_NV" },
  { 0x8656, "VERTEX_ATTRIB_ARRAY6_NV" },
  { 0x8657, "VERTEX_ATTRIB_ARRAY7_NV" },
  { 0x8658, "VERTEX_ATTRIB_ARRAY8_NV" },
  { 0x8659, "VERTEX_ATTRIB_ARRAY9_NV" },
  { 0x865a, "VERTEX_ATTRIB_ARRAY10_NV" },
  { 0x865b, "VERTEX_ATTRIB_ARRAY11_NV" },
  { 0x865c, "VERTEX_ATTRIB_ARRAY12_NV" },
  { 0x865d, "VERTEX_ATTRIB_ARRAY13_NV" },
  { 0x865e, "VERTEX_ATTRIB_ARRAY14_NV" },
  { 0x865f, "VERTEX_ATTRIB_ARRAY15_NV" },
  { 0x8660, "MAP1_VERTEX_ATTRIB0_4_NV" },
  { 0x8661, "MAP1_VERTEX_ATTRIB1_4_NV" },
  { 0x8662, "MAP1_VERTEX_ATTRIB2_4_NV" },
  { 0x8663, "MAP1_VERTEX_ATTRIB3_4_NV" },
  { 0x8664, "MAP1_VERTEX_ATTRIB4_4_NV" },
  { 0x8665, "MAP1_VERTEX_ATTRIB5_4_NV" },
  { 0x8666, "MAP1_VERTEX_ATTRIB6_4_NV" },
  { 0x8667, "MAP1_VERTEX_ATTRIB7_4_NV" },
  { 0x8668, "MAP1_VERTEX_ATTRIB8_4_NV" },
  { 0x8669, "MAP1_VERTEX_ATTRIB9_4_NV" },
  { 0x866a, "MAP1_VERTEX_ATTRIB10_4_NV" },
  { 0x866b, "MAP1_VERTEX_ATTRIB11_4_NV" },
  { 0x866c, "MAP1_VERTEX_ATTRIB12_4_NV" },
  { 0x866d, "MAP1_VERTEX_ATTRIB13_4_NV" },
  { 0x866e, "MAP1_VERTEX_ATTRIB14_4_NV" },
  { 0x866f, "MAP1_VERTEX_ATTRIB15_4_NV" },
  { 0x8670, "MAP2_VERTEX_ATTRIB0_4_NV" },
  { 0x8671, "MAP2_VERTEX_ATTRIB1_4_NV" },
  { 0x8672, "MAP2_VERTEX_ATTRIB2_4_NV" },
  { 0x8673, "MAP2_VERTEX_ATTRIB3_4_NV" },
  { 0x8674, "MAP2_VERTEX_ATTRIB4_4_NV" },
  { 0x8675, "MAP2_VERTEX_ATTRIB5_4_NV" },
  { 0x8676, "MAP2_VERTEX_ATTRIB6_4_NV" },
  { 0x8677, "PROGRAM_BINDING_ARB" },
  { 0x8678, "MAP2_VERTEX_ATTRIB8_4_NV" },
  { 0x8679, "MAP2_VERTEX_ATTRIB9_4_NV" },
  { 0x867a, "MAP2_VERTEX_ATTRIB10_4_NV" },
  { 0x867b, "MAP2_VERTEX_ATTRIB11_4_NV" },
  { 0x867c, "MAP2_VERTEX_ATTRIB12_4_NV" },
  { 0x867d, "MAP2_VERTEX_ATTRIB13_4_NV" },
  { 0x867e, "MAP2_VERTEX_ATTRIB14_4_NV" },
  { 0x867f, "MAP2_VERTEX_ATTRIB15_4_NV" },
  { 0x86a0, "TEXTURE_COMPRESSED_IMAGE_SIZE" },
  { 0x86a1, "TEXTURE_COMPRESSED" },
  { 0x86a2, "NUM_COMPRESSED_TEXTURE_FORMATS" },
  { 0x86a3, "COMPRESSED_TEXTURE_FORMATS" },
  { 0x86a4, "MAX_VERTEX_UNITS_ARB" },
  { 0x86a5, "ACTIVE_VERTEX_UNITS_ARB" },
  { 0x86a6, "WEIGHT_SUM_UNITY_ARB" },
  { 0x86a7, "VERTEX_BLEND_ARB" },
  { 0x86a8, "CURRENT_WEIGHT_ARB" },
  { 0x86a9, "WEIGHT_ARRAY_TYPE_ARB" },
  { 0x86aa, "WEIGHT_ARRAY_STRIDE_ARB" },
  { 0x86ab, "WEIGHT_ARRAY_SIZE_ARB" },
  { 0x86ac, "WEIGHT_ARRAY_POINTER_ARB" },
  { 0x86ad, "WEIGHT_ARRAY_ARB" },
  { 0x86ae, "DOT3_RGB" },
  { 0x86af, "DOT3_RGBA" },
  { 0x86b0, "COMPRESSED_RGB_FXT1_3DFX" },
  { 0x86b1, "COMPRESSED_RGBA_FXT1_3DFX" },
  { 0x86b2, "MULTISAMPLE_3DFX" },
  { 0x86b3, "SAMPLE_BUFFERS_3DFX" },
  { 0x86b4, "SAMPLES_3DFX" },
  { 0x8722, "MODELVIEW2_ARB" },
  { 0x8723, "MODELVIEW3_ARB" },
  { 0x8724, "MODELVIEW4_ARB" },
  { 0x8725, "MODELVIEW5_ARB" },
  { 0x8726, "MODELVIEW6_ARB" },
  { 0x8727, "MODELVIEW7_ARB" },
  { 0x8728, "MODELVIEW8_ARB" },
  { 0x8729, "MODELVIEW9_ARB" },
  { 0x872a, "MODELVIEW10_ARB" },
  { 0x872b, "MODELVIEW11_ARB" },
  { 0x872c, "MODELVIEW12_ARB" },
  { 0x872d, "MODELVIEW13_ARB" },
  { 0x872e, "MODELVIEW14_ARB" },
  { 0x872f, "MODELVIEW15_ARB" },
  { 0x8730, "MODELVIEW16_ARB" },
  { 0x8731, "MODELVIEW17_ARB" },
  { 0x8732, "MODELVIEW18_ARB" },
  { 0x8733, "MODELVIEW19_ARB" },
  { 0x8734, "MODELVIEW20_ARB" },
  { 0x8735, "MODELVIEW21_ARB" },
  { 0x8736, "MODELVIEW22_ARB" },
  { 0x8737, "MODELVIEW23_ARB" },
  { 0x8738, "MODELVIEW24_ARB" },
  { 0x8739, "MODELVIEW25_ARB" },
  { 0x873a, "MODELVIEW26_ARB" },
  { 0x873b, "MODELVIEW27_ARB" },
  { 0x873c, "MODELVIEW28_ARB" },
  { 0x873d, "MODELVIEW29_ARB" },
  { 0x873e, "MODELVIEW30_ARB" },
  { 0x873f, "MODELVIEW31_ARB" },
  { 0x8740, "DOT3_RGB_EXT" },
  { 0x8741, "DOT3_RGBA_EXT" },
  { 0x8742, "MIRROR_CLAMP_ATI" },
  { 0x8743, "MIRROR_CLAMP_TO_EDGE_ATI" },
  { 0x8744, "MODULATE_ADD_ATI" },
  { 0x8745, "MODULATE_SIGNED_ADD_ATI" },
  { 0x8746, "MODULATE_SUBTRACT_ATI" },
  { 0x8757, "YCBCR_MESA" },
  { 0x8758, "PACK_INVERT_MESA" },
  { 0x8764, "BUFFER_SIZE" },
  { 0x8765, "BUFFER_USAGE" },
  { 0x8775, "BUMP_ROT_MATRIX_ATI" },
  { 0x8776, "BUMP_ROT_MATRIX_SIZE_ATI" },
  { 0x8777, "BUMP_NUM_TEX_UNITS_ATI" },
  { 0x8778, "BUMP_TEX_UNITS_ATI" },
  { 0x8779, "DUDV_ATI" },
  { 0x877a, "DU8DV8_ATI" },
  { 0x877b, "BUMP_ENVMAP_ATI" },
  { 0x877c, "BUMP_TARGET_ATI" },
  { 0x87fb, "VBO_FREE_MEMORY_ATI" },
  { 0x87fc, "TEXTURE_FREE_MEMORY_ATI" },
  { 0x87fd, "RENDERBUFFER_FREE_MEMORY_ATI" },
  { 0x8800, "STENCIL_BACK_FUNC" },
  { 0x8801, "STENCIL_BACK_FAIL" },
  { 0x8802, "STENCIL_BACK_PASS_DEPTH_FAIL" },
  { 0x8803, "STENCIL_BACK_PASS_DEPTH_PASS" },
  { 0x8804, "FRAGMENT_PROGRAM_ARB" },
  { 0x8805, "PROGRAM_ALU_INSTRUCTIONS_ARB" },
  { 0x8806, "PROGRAM_TEX_INSTRUCTIONS_ARB" },
  { 0x8807, "PROGRAM_TEX_INDIRECTIONS_ARB" },
  { 0x8808, "PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB" },
  { 0x8809, "PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB" },
  { 0x880a, "PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB" },
  { 0x880b, "MAX_PROGRAM_ALU_INSTRUCTIONS_ARB" },
  { 0x880c, "MAX_PROGRAM_TEX_INSTRUCTIONS_ARB" },
  { 0x880d, "MAX_PROGRAM_TEX_INDIRECTIONS_ARB" },
  { 0x880e, "MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB" },
  { 0x880f, "MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB" },
  { 0x8810, "MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB" },
  { 0x8824, "MAX_DRAW_BUFFERS" },
  { 0x8825, "DRAW_BUFFER0" },
  { 0x8826, "DRAW_BUFFER1" },
  { 0x8827, "DRAW_BUFFER2" },
  { 0x8828, "DRAW_BUFFER3" },
  { 0x8829, "DRAW_BUFFER4" },
  { 0x882a, "DRAW_BUFFER5" },
  { 0x882b, "DRAW_BUFFER6" },
  { 0x882c, "DRAW_BUFFER7" },
  { 0x882d, "DRAW_BUFFER8" },
  { 0x882e, "DRAW_BUFFER9" },
  { 0x882f, "DRAW_BUFFER10" },
  { 0x8830, "DRAW_BUFFER11" },
  { 0x8831, "DRAW_BUFFER12" },
  { 0x8832, "DRAW_BUFFER13" },
  { 0x8833, "DRAW_BUFFER14" },
  { 0x8834, "DRAW_BUFFER15" },
  { 0x883d, "BLEND_EQUATION_ALPHA" },
  { 0x8840, "MATRIX_PALETTE_ARB" },
  { 0x8841, "MAX_MATRIX_PALETTE_STACK_DEPTH_ARB" },
  { 0x8842, "MAX_PALETTE_MATRICES_ARB" },
  { 0x8843, "CURRENT_PALETTE_MATRIX_ARB" },
  { 0x8844, "MATRIX_INDEX_ARRAY_ARB" },
  { 0x8845, "CURRENT_MATRIX_INDEX_ARB" },
  { 0x8846, "MATRIX_INDEX_ARRAY_SIZE_ARB" },
  { 0x8847, "MATRIX_INDEX_ARRAY_TYPE_ARB" },
  { 0x8848, "MATRIX_INDEX_ARRAY_STRIDE_ARB" },
  { 0x8849, "MATRIX_INDEX_ARRAY_POINTER_ARB" },
  { 0x884a, "TEXTURE_DEPTH_SIZE" },
  { 0x884b, "DEPTH_TEXTURE_MODE" },
  { 0x884c, "TEXTURE_COMPARE_MODE" },
  { 0x884d, "TEXTURE_COMPARE_FUNC" },
  { 0x884e, "COMPARE_R_TO_TEXTURE" },
  { 0x8861, "POINT_SPRITE" },
  { 0x8862, "COORD_REPLACE" },
  { 0x8863, "POINT_SPRITE_R_MODE_NV" },
  { 0x8864, "QUERY_COUNTER_BITS" },
  { 0x8865, "CURRENT_QUERY" },
  { 0x8866, "QUERY_RESULT" },
  { 0x8867, "QUERY_RESULT_AVAILABLE" },
  { 0x8868, "MAX_FRAGMENT_PROGRAM_LOCAL_PARAMETERS_NV" },
  { 0x8869, "MAX_VERTEX_ATTRIBS" },
  { 0x886a, "VERTEX_ATTRIB_ARRAY_NORMALIZED" },
  { 0x886e, "DEPTH_STENCIL_TO_RGBA_NV" },
  { 0x886f, "DEPTH_STENCIL_TO_BGRA_NV" },
  { 0x8870, "FRAGMENT_PROGRAM_NV" },
  { 0x8871, "MAX_TEXTURE_COORDS" },
  { 0x8872, "MAX_TEXTURE_IMAGE_UNITS" },
  { 0x8873, "FRAGMENT_PROGRAM_BINDING_NV" },
  { 0x8874, "PROGRAM_ERROR_STRING_ARB" },
  { 0x8875, "PROGRAM_FORMAT_ASCII_ARB" },
  { 0x8876, "PROGRAM_FORMAT_ARB" },
  { 0x888f, "TEXTURE_UNSIGNED_REMAP_MODE_NV" },
  { 0x8890, "DEPTH_BOUNDS_TEST_EXT" },
  { 0x8891, "DEPTH_BOUNDS_EXT" },
  { 0x8892, "ARRAY_BUFFER" },
  { 0x8893, "ELEMENT_ARRAY_BUFFER" },
  { 0x8894, "ARRAY_BUFFER_BINDING" },
  { 0x8895, "ELEMENT_ARRAY_BUFFER_BINDING" },
  { 0x8896, "VERTEX_ARRAY_BUFFER_BINDING" },
  { 0x8897, "NORMAL_ARRAY_BUFFER_BINDING" },
  { 0x8898, "COLOR_ARRAY_BUFFER_BINDING" },
  { 0x8899, "INDEX_ARRAY_BUFFER_BINDING" },
  { 0x889a, "TEXTURE_COORD_ARRAY_BUFFER_BINDING" },
  { 0x889b, "EDGE_FLAG_ARRAY_BUFFER_BINDING" },
  { 0x889c, "SECONDARY_COLOR_ARRAY_BUFFER_BINDING" },
  { 0x889d, "FOG_COORDINATE_ARRAY_BUFFER_BINDING" },
  { 0x889e, "WEIGHT_ARRAY_BUFFER_BINDING" },
  { 0x889f, "VERTEX_ATTRIB_ARRAY_BUFFER_BINDING" },
  { 0x88a0, "PROGRAM_INSTRUCTIONS_ARB" },
  { 0x88a1, "MAX_PROGRAM_INSTRUCTIONS_ARB" },
  { 0x88a2, "PROGRAM_NATIVE_INSTRUCTIONS_ARB" },
  { 0x88a3, "MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB" },
  { 0x88a4, "PROGRAM_TEMPORARIES_ARB" },
  { 0x88a5, "MAX_PROGRAM_TEMPORARIES_ARB" },
  { 0x88a6, "PROGRAM_NATIVE_TEMPORARIES_ARB" },
  { 0x88a7, "MAX_PROGRAM_NATIVE_TEMPORARIES_ARB" },
  { 0x88a8, "PROGRAM_PARAMETERS_ARB" },
  { 0x88a9, "MAX_PROGRAM_PARAMETERS_ARB" },
  { 0x88aa, "PROGRAM_NATIVE_PARAMETERS_ARB" },
  { 0x88ab, "MAX_PROGRAM_NATIVE_PARAMETERS_ARB" },
  { 0x88ac, "PROGRAM_ATTRIBS_ARB" },
  { 0x88ad, "MAX_PROGRAM_ATTRIBS_ARB" },
  { 0x88ae, "PROGRAM_NATIVE_ATTRIBS_ARB" },
  { 0x88af, "MAX_PROGRAM_NATIVE_ATTRIBS_ARB" },
  { 0x88b0, "PROGRAM_ADDRESS_REGISTERS_ARB" },
  { 0x88b1, "MAX_PROGRAM_ADDRESS_REGISTERS_ARB" },
  { 0x88b2, "PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB" },
  { 0x88b3, "MAX_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB" },
  { 0x88b4, "MAX_PROGRAM_LOCAL_PARAMETERS_ARB" },
  { 0x88b5, "MAX_PROGRAM_ENV_PARAMETERS_ARB" },
  { 0x88b6, "PROGRAM_UNDER_NATIVE_LIMITS_ARB" },
  { 0x88b7, "TRANSPOSE_CURRENT_MATRIX_ARB" },
  { 0x88b8, "READ_ONLY" },
  { 0x88b9, "WRITE_ONLY" },
  { 0x88ba, "READ_WRITE" },
  { 0x88bb, "BUFFER_ACCESS" },
  { 0x88bc, "BUFFER_MAPPED" },
  { 0x88bd, "BUFFER_MAP_POINTER" },
  { 0x88bf, "TIME_ELAPSED" },
  { 0x88c0, "MATRIX0_ARB" },
  { 0x88c1, "MATRIX1_ARB" },
  { 0x88c2, "MATRIX2_ARB" },
  { 0x88c3, "MATRIX3_ARB" },
  { 0x88c4, "MATRIX4_ARB" },
  { 0x88c5, "MATRIX5_ARB" },
  { 0x88c6, "MATRIX6_ARB" },
  { 0x88c7, "MATRIX7_ARB" },
  { 0x88c8, "MATRIX8_ARB" },
  { 0x88c9, "MATRIX9_ARB" },
  { 0x88ca, "MATRIX10_ARB" },
  { 0x88cb, "MATRIX11_ARB" },
  { 0x88cc, "MATRIX12_ARB" },
  { 0x88cd, "MATRIX13_ARB" },
  { 0x88ce, "MATRIX14_ARB" },
  { 0x88cf, "MATRIX15_ARB" },
  { 0x88d0, "MATRIX16_ARB" },
  { 0x88d1, "MATRIX17_ARB" },
  { 0x88d2, "MATRIX18_ARB" },
  { 0x88d3, "MATRIX19_ARB" },
  { 0x88d4, "MATRIX20_ARB" },
  { 0x88d5, "MATRIX21_ARB" },
  { 0x88d6, "MATRIX22_ARB" },
  { 0x88d7, "MATRIX23_ARB" },
  { 0x88d8, "MATRIX24_ARB" },
  { 0x88d9, "MATRIX25_ARB" },
  { 0x88da, "MATRIX26_ARB" },
  { 0x88db, "MATRIX27_ARB" },
  { 0x88dc, "MATRIX28_ARB" },
  { 0x88dd, "MATRIX29_ARB" },
  { 0x88de, "MATRIX30_ARB" },
  { 0x88df, "MATRIX31_ARB" },
  { 0x88e0, "STREAM_DRAW" },
  { 0x88e1, "STREAM_READ" },
  { 0x88e2, "STREAM_COPY" },
  { 0x88e4, "STATIC_DRAW" },
  { 0x88e5, "STATIC_READ" },
  { 0x88e6, "STATIC_COPY" },
  { 0x88e8, "DYNAMIC_DRAW" },
  { 0x88e9, "DYNAMIC_READ" },
  { 0x88ea, "DYNAMIC_COPY" },
  { 0x88eb, "PIXEL_PACK_BUFFER" },
  { 0x88ec, "PIXEL_UNPACK_BUFFER" },
  { 0x88ed, "PIXEL_PACK_BUFFER_BINDING" },
  { 0x88ef, "PIXEL_UNPACK_BUFFER_BINDING" },
  { 0x88f4, "MAX_PROGRAM_EXEC_INSTRUCTIONS_NV" },
  { 0x88f5, "MAX_PROGRAM_CALL_DEPTH_NV" },
  { 0x88f6, "MAX_PROGRAM_IF_DEPTH_NV" },
  { 0x88f7, "MAX_PROGRAM_LOOP_DEPTH_NV" },
  { 0x88f8, "MAX_PROGRAM_LOOP_COUNT_NV" },
  { 0x8910, "STENCIL_TEST_TWO_SIDE_EXT" },
  { 0x8911, "ACTIVE_STENCIL_FACE_EXT" },
  { 0x8912, "MIRROR_CLAMP_TO_BORDER_EXT" },
  { 0x8914, "SAMPLES_PASSED" },
  { 0x8920, "FRAGMENT_SHADER_ATI" },
  { 0x8921, "REG_0_ATI" },
  { 0x8922, "REG_1_ATI" },
  { 0x8923, "REG_2_ATI" },
  { 0x8924, "REG_3_ATI" },
  { 0x8925, "REG_4_ATI" },
  { 0x8926, "REG_5_ATI" },
  { 0x8927, "REG_6_ATI" },
  { 0x8928, "REG_7_ATI" },
  { 0x8929, "REG_8_ATI" },
  { 0x892a, "REG_9_ATI" },
  { 0x892b, "REG_10_ATI" },
  { 0x892c, "REG_11_ATI" },
  { 0x892d, "REG_12_ATI" },
  { 0x892e, "REG_13_ATI" },
  { 0x892f, "REG_14_ATI" },
  { 0x8930, "REG_15_ATI" },
  { 0x8931, "REG_16_ATI" },
  { 0x8932, "REG_17_ATI" },
  { 0x8933, "REG_18_ATI" },
  { 0x8934, "REG_19_ATI" },
  { 0x8935, "REG_20_ATI" },
  { 0x8936, "REG_21_ATI" },
  { 0x8937, "REG_22_ATI" },
  { 0x8938, "REG_23_ATI" },
  { 0x8939, "REG_24_ATI" },
  { 0x893a, "REG_25_ATI" },
  { 0x893b, "REG_26_ATI" },
  { 0x893c, "REG_27_ATI" },
  { 0x893d, "REG_28_ATI" },
  { 0x893e, "REG_29_ATI" },
  { 0x893f, "REG_30_ATI" },
  { 0x8940, "REG_31_ATI" },
  { 0x8941, "CON_0_ATI" },
  { 0x8942, "CON_1_ATI" },
  { 0x8943, "CON_2_ATI" },
  { 0x8944, "CON_3_ATI" },
  { 0x8945, "CON_4_ATI" },
  { 0x8946, "CON_5_ATI" },
  { 0x8947, "CON_6_ATI" },
  { 0x8948, "CON_7_ATI" },
  { 0x8949, "CON_8_ATI" },
  { 0x894a, "CON_9_ATI" },
  { 0x894b, "CON_10_ATI" },
  { 0x894c, "CON_11_ATI" },
  { 0x894d, "CON_12_ATI" },
  { 0x894e, "CON_13_ATI" },
  { 0x894f, "CON_14_ATI" },
  { 0x8950, "CON_15_ATI" },
  { 0x8951, "CON_16_ATI" },
  { 0x8952, "CON_17_ATI" },
  { 0x8953, "CON_18_ATI" },
  { 0x8954, "CON_19_ATI" },
  { 0x8955, "CON_20_ATI" },
  { 0x8956, "CON_21_ATI" },
  { 0x8957, "CON_22_ATI" },
  { 0x8958, "CON_23_ATI" },
  { 0x8959, "CON_24_ATI" },
  { 0x895a, "CON_25_ATI" },
  { 0x895b, "CON_26_ATI" },
  { 0x895c, "CON_27_ATI" },
  { 0x895d, "CON_28_ATI" },
  { 0x895e, "CON_29_ATI" },
  { 0x895f, "CON_30_ATI" },
  { 0x8960, "CON_31_ATI" },
  { 0x8961, "MOV_ATI" },
  { 0x8963, "ADD_ATI" },
  { 0x8964, "MUL_ATI" },
  { 0x8965, "SUB_ATI" },
  { 0x8966, "DOT3_ATI" },
  { 0x8967, "DOT4_ATI" },
  { 0x8968, "MAD_ATI" },
  { 0x8969, "LERP_ATI" },
  { 0x896a, "CND_ATI" },
  { 0x896b, "CND0_ATI" },
  { 0x896c, "DOT2_ADD_ATI" },
  { 0x896d, "SECONDARY_INTERPOLATOR_ATI" },
  { 0x896e, "NUM_FRAGMENT_REGISTERS_ATI" },
  { 0x896f, "NUM_FRAGMENT_CONSTANTS_ATI" },
  { 0x8970, "NUM_PASSES_ATI" },
  { 0x8971, "NUM_INSTRUCTIONS_PER_PASS_ATI" },
  { 0x8972, "NUM_INSTRUCTIONS_TOTAL_ATI" },
  { 0x8973, "NUM_INPUT_INTERPOLATOR_COMPONENTS_ATI" },
  { 0x8974, "NUM_LOOPBACK_COMPONENTS_ATI" },
  { 0x8975, "COLOR_ALPHA_PAIRING_ATI" },
  { 0x8976, "SWIZZLE_STR_ATI" },
  { 0x8977, "SWIZZLE_STQ_ATI" },
  { 0x8978, "SWIZZLE_STR_DR_ATI" },
  { 0x8979, "SWIZZLE_STQ_DQ_ATI" },
  { 0x897a, "SWIZZLE_STRQ_ATI" },
  { 0x897b, "SWIZZLE_STRQ_DQ_ATI" },
  { 0x8a12, "BUFFER_SERIALIZED_MODIFY_APPLE" },
  { 0x8a13, "BUFFER_FLUSHING_UNMAP_APPLE" },
  { 0x8a48, "TEXTURE_SRGB_DECODE_EXT" },
  { 0x8a49, "DECODE_EXT" },
  { 0x8a4a, "SKIP_DECODE_EXT" },
  { 0x8a52, "FRAGMENT_SHADER_DISCARDS_SAMPLES_EXT" },
  { 0x8b30, "FRAGMENT_SHADER" },
  { 0x8b31, "VERTEX_SHADER" },
  { 0x8b40, "PROGRAM_OBJECT_ARB" },
  { 0x8b48, "SHADER_OBJECT_ARB" },
  { 0x8b49, "MAX_FRAGMENT_UNIFORM_COMPONENTS" },
  { 0x8b4a, "MAX_VERTEX_UNIFORM_COMPONENTS" },
  { 0x8b4b, "MAX_VARYING_FLOATS" },
  { 0x8b4c, "MAX_VERTEX_TEXTURE_IMAGE_UNITS" },
  { 0x8b4d, "MAX_COMBINED_TEXTURE_IMAGE_UNITS" },
  { 0x8b4e, "OBJECT_TYPE_ARB" },
  { 0x8b4f, "SHADER_TYPE" },
  { 0x8b50, "FLOAT_VEC2" },
  { 0x8b51, "FLOAT_VEC3" },
  { 0x8b52, "FLOAT_VEC4" },
  { 0x8b53, "INT_VEC2" },
  { 0x8b54, "INT_VEC3" },
  { 0x8b55, "INT_VEC4" },
  { 0x8b56, "BOOL" },
  { 0x8b57, "BOOL_VEC2" },
  { 0x8b58, "BOOL_VEC3" },
  { 0x8b59, "BOOL_VEC4" },
  { 0x8b5a, "FLOAT_MAT2" },
  { 0x8b5b, "FLOAT_MAT3" },
  { 0x8b5c, "FLOAT_MAT4" },
  { 0x8b5d, "SAMPLER_1D" },
  { 0x8b5e, "SAMPLER_2D" },
  { 0x8b5f, "SAMPLER_3D" },
  { 0x8b60, "SAMPLER_CUBE" },
  { 0x8b61, "SAMPLER_1D_SHADOW" },
  { 0x8b62, "SAMPLER_2D_SHADOW" },
  { 0x8b65, "FLOAT_MAT2x3" },
  { 0x8b66, "FLOAT_MAT2x4" },
  { 0x8b67, "FLOAT_MAT3x2" },
  { 0x8b68, "FLOAT_MAT3x4" },
  { 0x8b69, "FLOAT_MAT4x2" },
  { 0x8b6a, "FLOAT_MAT4x3" },
  { 0x8b80, "DELETE_STATUS" },
  { 0x8b81, "COMPILE_STATUS" },
  { 0x8b82, "LINK_STATUS" },
  { 0x8b83, "VALIDATE_STATUS" },
  { 0x8b84, "INFO_LOG_LENGTH" },
  { 0x8b85, "ATTACHED_SHADERS" },
  { 0x8b86, "ACTIVE_UNIFORMS" },
  { 0x8b87, "ACTIVE_UNIFORM_MAX_LENGTH" },
  { 0x8b88, "SHADER_SOURCE_LENGTH" },
  { 0x8b89, "ACTIVE_ATTRIBUTES" },
  { 0x8b8a, "ACTIVE_ATTRIBUTE_MAX_LENGTH" },
  { 0x8b8b, "FRAGMENT_SHADER_DERIVATIVE_HINT" },
  { 0x8b8c, "SHADING_LANGUAGE_VERSION" },
  { 0x8b8d, "CURRENT_PROGRAM" },
  { 0x8b90, "PALETTE4_RGB8_OES" },
  { 0x8b91, "PALETTE4_RGBA8_OES" },
  { 0x8b92, "PALETTE4_R5_G6_B5_OES" },
  { 0x8b93, "PALETTE4_RGBA4_OES" },
  { 0x8b94, "PALETTE4_RGB5_A1_OES" },
  { 0x8b95, "PALETTE8_RGB8_OES" },
  { 0x8b96, "PALETTE8_RGBA8_OES" },
  { 0x8b97, "PALETTE8_R5_G6_B5_OES" },
  { 0x8b98, "PALETTE8_RGBA4_OES" },
  { 0x8b99, "PALETTE8_RGB5_A1_OES" },
  { 0x8b9a, "IMPLEMENTATION_COLOR_READ_TYPE_OES" },
  { 0x8b9b, "IMPLEMENTATION_COLOR_READ_FORMAT_OES" },
  { 0x8c2f, "ANY_SAMPLES_PASSED" },
  { 0x8c3a, "R11F_G11F_B10F_EXT" },
  { 0x8c3b, "UNSIGNED_INT_10F_11F_11F_REV" },
  { 0x8c3c, "RGBA_SIGNED_COMPONENTS_EXT" },
  { 0x8c40, "SRGB" },
  { 0x8c41, "SRGB8" },
  { 0x8c42, "SRGB_ALPHA" },
  { 0x8c43, "SRGB8_ALPHA8" },
  { 0x8c44, "SLUMINANCE_ALPHA" },
  { 0x8c45, "SLUMINANCE8_ALPHA8" },
  { 0x8c46, "SLUMINANCE" },
  { 0x8c47, "SLUMINANCE8" },
  { 0x8c48, "COMPRESSED_SRGB" },
  { 0x8c49, "COMPRESSED_SRGB_ALPHA" },
  { 0x8c4a, "COMPRESSED_SLUMINANCE" },
  { 0x8c4b, "COMPRESSED_SLUMINANCE_ALPHA" },
  { 0x8ca0, "POINT_SPRITE_COORD_ORIGIN" },
  { 0x8ca1, "LOWER_LEFT" },
  { 0x8ca2, "UPPER_LEFT" },
  { 0x8ca3, "STENCIL_BACK_REF" },
  { 0x8ca4, "STENCIL_BACK_VALUE_MASK" },
  { 0x8ca5, "STENCIL_BACK_WRITEMASK" },
  { 0x8db9, "FRAMEBUFFER_SRGB_EXT" },
  { 0x8dba, "FRAMEBUFFER_SRGB_CAPABLE_EXT" },
  { 0x8e17, "QUERY_WAIT_INVERTED" },
  { 0x8e18, "QUERY_NO_WAIT_INVERTED" },
  { 0x8e19, "QUERY_BY_REGION_WAIT_INVERTED" },
  { 0x8e1a, "QUERY_BY_REGION_NO_WAIT_INVERTED" },
  { 0x8e1b, "POLYGON_OFFSET_CLAMP_EXT" },
  { 0x8e28, "TIMESTAMP" },
  { 0x8e50, "SAMPLE_LOCATION_ARB" },
  { 0x8e70, "MAX_TRANSFORM_FEEDBACK_BUFFERS" },
  { 0x8e71, "MAX_VERTEX_STREAMS" },
  { 0x8f93, "RGBA_SNORM" },
  { 0x8f97, "RGBA8_SNORM" },
  { 0x8f9c, "SIGNED_NORMALIZED" },
  { 0x9047, "GPU_MEMORY_INFO_DEDICATED_VIDMEM_NVX" },
  { 0x9048, "GPU_MEMORY_INFO_TOTAL_AVAILABLE_MEMORY_NVX" },
  { 0x9049, "GPU_MEMORY_INFO_CURRENT_AVAILABLE_VIDMEM_NVX" },
  { 0x904a, "GPU_MEMORY_INFO_EVICTION_COUNT_NVX" },
  { 0x904b, "GPU_MEMORY_INFO_EVICTED_MEMORY_NVX" },
  { 0x90bc, "MIN_MAP_BUFFER_ALIGNMENT" },
  { 0x9160, "EXTERNAL_VIRTUAL_MEMORY_BUFFER_AMD" },
  { 0x9192, "QUERY_BUFFER" },
  { 0x9193, "QUERY_BUFFER_BINDING" },
  { 0x9194, "QUERY_RESULT_NO_WAIT" },
  { 0x91b0, "MAX_SHADER_COMPILER_THREADS_ARB" },
  { 0x91b1, "COMPLETION_STATUS_ARB" },
  { 0x91b2, "RENDERBUFFER_STORAGE_SAMPLES_AMD" },
  { 0x91b3, "MAX_COLOR_FRAMEBUFFER_SAMPLES_AMD" },
  { 0x91b4, "MAX_COLOR_FRAMEBUFFER_STORAGE_SAMPLES_AMD" },
  { 0x91b5, "MAX_DEPTH_STENCIL_FRAMEBUFFER_SAMPLES_AMD" },
  { 0x91b6, "NUM_SUPPORTED_MULTISAMPLE_MODES_AMD" },
  { 0x91b7, "SUPPORTED_MULTISAMPLE_MODES_AMD" },
  { 0x9285, "BLEND_ADVANCED_COHERENT_KHR" },
  { 0x9294, "MULTIPLY_KHR" },
  { 0x9295, "SCREEN_KHR" },
  { 0x9296, "OVERLAY_KHR" },
  { 0x9297, "DARKEN_KHR" },
  { 0x9298, "LIGHTEN_KHR" },
  { 0x9299, "COLORDODGE_KHR" },
  { 0x929a, "COLORBURN_KHR" },
  { 0x929b, "HARDLIGHT_KHR" },
  { 0x929c, "SOFTLIGHT_KHR" },
  { 0x929e, "DIFFERENCE_KHR" },
  { 0x92a0, "EXCLUSION_KHR" },
  { 0x92ad, "HSL_HUE_KHR" },
  { 0x92ae, "HSL_SATURATION_KHR" },
  { 0x92af, "HSL_COLOR_KHR" },
  { 0x92b0, "HSL_LUMINOSITY_KHR" },
  { 0x92be, "PRIMITIVE_BOUNDING_BOX_ARB" },
  { 0x933c, "FILL_RECTANGLE_NV" },
  { 0x933d, "SAMPLE_LOCATION_SUBPIXEL_BITS_ARB" },
  { 0x933e, "SAMPLE_LOCATION_PIXEL_GRID_WIDTH_ARB" },
  { 0x933f, "SAMPLE_LOCATION_PIXEL_GRID_HEIGHT_ARB" },
  { 0x9340, "PROGRAMMABLE_SAMPLE_LOCATION_TABLE_SIZE_ARB" },
  { 0x9341, "PROGRAMMABLE_SAMPLE_LOCATION_ARB" },
  { 0x9342, "FRAMEBUFFER_PROGRAMMABLE_SAMPLE_LOCATIONS_ARB" },
  { 0x9343, "FRAMEBUFFER_SAMPLE_LOCATION_PIXEL_GRID_ARB" },
  { 0x9346, "CONSERVATIVE_RASTERIZATION_NV" },
  { 0x9347, "SUBPIXEL_PRECISION_BIAS_X_BITS_NV" },
  { 0x9348, "SUBPIXEL_PRECISION_BIAS_Y_BITS_NV" },
  { 0x9349, "MAX_SUBPIXEL_PRECISION_BIAS_BITS_NV" },
  { 0x9379, "CONSERVATIVE_RASTER_DILATE_NV" },
  { 0x937a, "CONSERVATIVE_RASTER_DILATE_RANGE_NV" },
  { 0x937b, "CONSERVATIVE_RASTER_DILATE_GRANULARITY_NV" },
  { 0x9381, "MULTISAMPLE_LINE_WIDTH_RANGE_ARB" },
  { 0x9382, "MULTISAMPLE_LINE_WIDTH_GRANULARITY_ARB" },
  { 0x954d, "CONSERVATIVE_RASTER_MODE_NV" },
  { 0x954e, "CONSERVATIVE_RASTER_MODE_POST_SNAP_NV" },
  { 0x954f, "CONSERVATIVE_RASTER_MODE_PRE_SNAP_TRIANGLES_NV" },
  { 0x10000, "EVAL_BIT" },
  { 0x19262, "RASTER_POSITION_UNCLIPPED_IBM" },
  { 0x20000, "LIST_BIT" },
  { 0x40000, "TEXTURE_BIT" },
  { 0x80000, "SCISSOR_BIT" },
  { 0xfffff, "ALL_ATTRIB_BITS" },
  { 0x20000000, "MULTISAMPLE_BIT" },
  { 0xffffffff, "CLIENT_ALL_ATTRIB_BITS" },
    { 0, NULL }
};
